#ifndef FRAMES_H_
#define FRAMES_H_

#include <stdlib.h>
#include <string.h>
#include <stddef.h>

#include "nob.h"

#define ARENA_IMPLEMENTATION
#include "arena.h"

// == Frames data-structure ===================================================

typedef void (*FrameDraw)(float time, float scale);
typedef void (*FrameTick)(float time, void *state);

#define FRAMES_EMPTY_TICK  NULL
#define FRAMES_EMPTY_STATE 0

typedef struct {
  FrameDraw draw;
  FrameTick tick;

  void*  state;
  size_t state_size;
} Frame;

typedef struct {
  Frame* items;
  size_t count;
  size_t capacity;

  size_t cursor;
  float  time;

  Arena states;
} Frames;

void frames_next(Frames *fs)
{
  assert(fs->cursor < fs->count);
  if (fs->cursor == fs->count - 1) return;

  fs->cursor++;
  fs->time = 0;
}

void frames_prev(Frames *fs)
{
  assert(fs->cursor < fs->count);
  if (fs->cursor == 0) return;

  fs->cursor--;
  fs->time = 0;
}

void frames_first(Frames *fs)
{
  assert(fs->cursor < fs->count);
  if (fs->cursor == 0) return;

  fs->cursor = 0;
  fs->time = 0;
}

void frames_last(Frames *fs)
{
  assert(fs->cursor < fs->count);
  if (fs->count == fs->count - 1) return;

  fs->cursor = fs->count - 1;
  fs->time = 0;
}

void frames_draw(Frames *fs, float scale)
{
  assert(fs->cursor < fs->count);
  if (fs->count == 0) return;

  Frame f = fs->items[fs->cursor];
  assert(f.draw != NULL);

  BeginDrawing();
  f.draw(fs->time, scale);
  EndDrawing();
}

void frames_tick(Frames *fs, float dt)
{
  assert(fs->cursor < fs->count);

  fs->time += dt;
  if (fs->time < 0) fs->time = 0;

  if (fs->count == 0) return;
  Frame f = fs->items[fs->cursor];
  if (f.tick != FRAMES_EMPTY_TICK) f.tick(fs->time, f.state);
}

void frames_append(Frames *fs, FrameDraw draw, FrameTick tick,
                   size_t state_size)
{
  void *state = NULL;
  if (state_size > 0) {
    state = arena_alloc(&fs->states, state_size);
    memset(state, 0, state_size);
  }

  Frame f = {
    .draw       = draw,
    .tick       = tick,
    .state      = state,
    .state_size = state_size,
  };
  nob_da_append(fs, f);
}

typedef void (*FramesLoad)(Frames *fs);

void frames_load(Frames *fs, FramesLoad load)
{
  fs->count = 0;
  arena_reset(&fs->states);

  load(fs);

  if (fs->cursor >= fs->count) fs->cursor = fs->count - 1;
}

// == Text layout =============================================================

typedef struct {
  const char* text;
  Color       color;

  size_t width;
  size_t x_offset;
} Span;

typedef struct {
  Span*  items;
  size_t count;
  size_t capacity;

  size_t width;
  int    font_size;
} Line;

void line_append(Line *line, const char *text, Color color)
{
  size_t text_width = MeasureText(text, line->font_size);
  Span le = {
    .text     = text,
    .color    = color,
    .width    = text_width,
    .x_offset = line->width,
  };

  nob_da_append(line, le);
  line->width += text_width;
}

Line line(const char *text, Color color, int font_size)
{
  Line line = {0};
  line.font_size = font_size;

  line_append(&line, text, color);
  return line;
}

void line_draw(Line *line, int x, int y)
{
  for (size_t i = 0; i < line->count; i++) {
    Span le = line->items[i];
    DrawText(le.text, x + le.x_offset, y, line->font_size, le.color);
  }
}

void line_draw_centered(Line *line, int y)
{
  int x = (GetScreenWidth() - line->width)/2;
  line_draw(line, x, y);
}

#endif // FRAMES_H_
